﻿Imports Microsoft.VisualBasic
Imports System
Namespace LogisticGradient
    Friend Class LogisticGradientProgram
        Shared Sub Main(ByVal args() As String)
            Console.WriteLine(VBCrLf & "Begin Logistic Regression (binary) Classification demo")
            Console.WriteLine("Goal is to demonstrate training using gradient descent")

            Dim numFeatures As Integer = 8 ' synthetic data
            Dim numRows As Integer = 10000
            Dim seed As Integer = 1

            Console.WriteLine(VBCrLf, "Generating ", numRows, " artificial data items with ", numFeatures, " features")
            Dim allData()() As Double = MakeAllData(numFeatures, numRows, seed)

            Console.WriteLine("Creating train (80%) and test (20%) matrices")
            Dim trainData()() As Double
            Dim testData()() As Double
            MakeTrainTest(allData, 0, trainData, testData)
            Console.WriteLine("Done")


            Console.WriteLine(VBCrLf & "Training data: " & VBCrLf)
            ShowData(trainData, 3, 2, True)

            Console.WriteLine(VBCrLf & "Test data: " & VBCrLf)
            ShowData(testData, 3, 2, True)


            Console.WriteLine("Creating LR binary classifier")
            Dim lc As New LogisticClassifier(numFeatures)

            Dim maxEpochs As Integer = 1000
            Console.WriteLine("Setting maxEpochs = " & maxEpochs)
            Dim alpha As Double = 0.01
            Console.WriteLine("Setting learning rate = " & alpha.ToString("F2"))

            Console.WriteLine(VBCrLf & "Starting training using (stochastic) gradient descent")
            Dim weights() As Double = lc.Train(trainData, maxEpochs, alpha)
            Console.WriteLine("Training complete")

            Console.WriteLine(VBCrLf & "Best weights found:")
            ShowVector(weights, 4, True)

            Dim trainAcc As Double = lc.Accuracy(trainData, weights)
            Console.WriteLine("Prediction accuracy on training data = " & trainAcc.ToString("F4"))

            Dim testAcc As Double = lc.Accuracy(testData, weights)
            Console.WriteLine("Prediction accuracy on test data = " & testAcc.ToString("F4"))

            Console.WriteLine(VBCrLf & "End LR binary classification demo" & VBCrLf)
            Console.ReadLine()
        End Sub

        Private Shared Function MakeAllData(ByVal numFeatures As Integer, ByVal numRows As Integer, ByVal seed As Integer) As Double()()
            Dim rnd As New Random(seed)
            Dim weights(numFeatures) As Double ' inc. b0
            For i = 0 To weights.Length - 1
                weights(i) = 20.0 * rnd.NextDouble() - 10.0 ' [-10.0 to +10.0]
            Next i

            Dim result(numRows - 1)() As Double ' allocate matrix
            For i = 0 To numRows - 1
                result(i) = New Double(numFeatures) {} ' Y in last column
            Next i

            For i = 0 To numRows - 1 ' for each row
                Dim z As Double = weights(0) ' the b0
                For j = 0 To numFeatures - 1 ' each feature / column except last
                    Dim x As Double = 20.0 * rnd.NextDouble() - 10.0 ' random X in [10.0, +10.0]
                    result(i)(j) = x ' store x
                    Dim wx As Double = x * weights(j + 1) ' weight * x
                    z += wx ' accumulate to get Y
                Next j
                Dim y As Double = 1.0 / (1.0 + Math.Exp(-z))
                If y > 0.55 Then ' slight bias towards 0
                    result(i)(numFeatures) = 1.0 ' store y in last column
                Else
                    result(i)(numFeatures) = 0.0
                End If
            Next i
            Console.WriteLine("Data generation weights:")
            ShowVector(weights, 4, True)

            Return result
        End Function ' MakeAllData

        Private Shared Sub MakeTrainTest(ByVal allData()() As Double, ByVal seed As Integer, <System.Runtime.InteropServices.Out()> ByRef trainData()() As Double, <System.Runtime.InteropServices.Out()> ByRef testData()() As Double)
            Dim rnd As New Random(seed)
            Dim totRows As Integer = allData.Length
            Dim numTrainRows As Integer = CInt(Fix(totRows * 0.8)) ' 80% hard-coded
            Dim numTestRows As Integer = totRows - numTrainRows
            trainData = New Double(numTrainRows - 1)() {}
            testData = New Double(numTestRows - 1)() {}

            Dim copy(allData.Length - 1)() As Double ' ref copy of all data
            For i = 0 To copy.Length - 1
                copy(i) = allData(i)
            Next i

            For i = 0 To copy.Length - 1 ' scramble order
                Dim r As Integer = rnd.Next(i, copy.Length) ' use Fisher-Yates
                Dim tmp() As Double = copy(r)
                copy(r) = copy(i)
                copy(i) = tmp
            Next i
            For i = 0 To numTrainRows - 1
                trainData(i) = copy(i)
            Next i

            For i = 0 To numTestRows - 1
                testData(i) = copy(i + numTrainRows)
            Next i
        End Sub ' MakeTrainTest


        Public Shared Sub ShowData(ByVal data()() As Double, ByVal numRows As Integer, ByVal decimals As Integer, ByVal indices As Boolean)
            Dim len As Integer = data.Length.ToString().Length
            For i = 0 To numRows - 1
                If indices = True Then
                    Console.Write("[" & i.ToString().PadLeft(len) & "]  ")
                End If
                For j As Integer = 0 To data(i).Length - 1
                    Dim v As Double = data(i)(j)
                    If v >= 0.0 Then
                        Console.Write(" ") ' '+'
                    End If
                    Console.Write(v.ToString("F" & decimals) & "  ")
                Next j
                Console.WriteLine("")
            Next i
            Console.WriteLine(". . .")
            Dim lastRow As Integer = data.Length - 1
            If indices = True Then
                Console.Write("[" & lastRow.ToString().PadLeft(len) & "]  ")
            End If
            For j = 0 To data(lastRow).Length - 1
                Dim v As Double = data(lastRow)(j)
                If v >= 0.0 Then
                    Console.Write(" ") ' '+'
                End If
                Console.Write(v.ToString("F" & decimals) & "  ")
            Next j
            Console.WriteLine(VBCrLf)
        End Sub

        Private Shared Sub ShowVector(ByVal vector() As Double, ByVal decimals As Integer, ByVal newLine As Boolean)
            For i = 0 To vector.Length - 1
                Console.Write(vector(i).ToString("F" & decimals) & " ")
            Next i
            Console.WriteLine("")
            If newLine = True Then
                Console.WriteLine("")
            End If
        End Sub

    End Class ' Program

    Public Class LogisticClassifier
        Private numFeatures As Integer ' number of x variables aka features
        Private weights() As Double ' b0 = constant
        Private rnd As Random

        Public Sub New(ByVal numFeatures As Integer)
            Me.numFeatures = numFeatures
            Me.weights = New Double(numFeatures) {} ' [0] = b0 constant
            Me.rnd = New Random(0)
            'For i = 0 To weights.Length - 1 ' not necessary
            '    weights(i) = 0.01 * rnd.NextDouble() ' [0.00, 0.01)
            'Next i
        End Sub

        Public Function Train(ByVal trainData()() As Double, ByVal maxEpochs As Integer, ByVal alpha As Double) As Double()
            ' alpha is the learning rate
            Dim epoch As Integer = 0
            Dim sequence(trainData.Length - 1) As Integer ' random order
            For i = 0 To sequence.Length - 1
                sequence(i) = i
            Next i

            Do While epoch < maxEpochs
                epoch += 1

                If epoch Mod 100 = 0 AndAlso epoch <> maxEpochs Then
                    Dim mse As Double = [Error](trainData, weights)
                    Console.Write("epoch = " & epoch)
                    Console.WriteLine("  error = " & mse.ToString("F4"))
                End If

                Shuffle(sequence) ' process data in random order

                ' stochastic/online/incremental approach
                For ti = 0 To trainData.Length - 1
                    Dim i As Integer = sequence(ti)
                    Dim computed As Double = ComputeOutput(trainData(i), weights)
                    Dim targetIndex As Integer = trainData(i).Length - 1
                    Dim target As Double = trainData(i)(targetIndex)

                    weights(0) += alpha * (target - computed) * 1 ' the b0 weight has a dummy 1 input
                    'weights(0) += alpha * (target - computed) * computed * (1 -computed) * 1 ' alt. form
                    For j = 1 To weights.Length - 1
                        weights(j) += alpha * (target - computed) * trainData(i)(j - 1)
                    Next j
                    'weights(j) += alpha * (target - computed) * computed * (1 - computed) * trainData(i)(j - 1) ' alt. form
                Next ti

                ' batch/offline approach
                Dim accumulatedGradients(weights.Length - 1) As Double ' one acc for each weight

                'For i As Integer = 0 To trainData.Length - 1 ' accumulate
                '    Dim computed As Double = ComputeOutput(trainData(i), weights) ' no need to shuffle order
                '    Dim targetIndex As Integer = trainData(i).Length - 1
                '    Dim target As Double = trainData(i)(targetIndex)
                '    accumulatedGradients(0) += (target - computed) * 1 ' for b0
                '    For j As Integer = 1 To weights.Length - 1
                '        accumulatedGradients(j) += (target - computed) * trainData(i)(j - 1)
                '    Next j
                'Next i

                'For j As Integer = 0 To weights.Length - 1 ' update
                '    weights(j) += alpha * accumulatedGradients(j)
                'Next j

            Loop ' while
            Return Me.weights ' by ref is somewhat risky
        End Function ' Train

        Private Sub Shuffle(ByVal sequence() As Integer)
            For i = 0 To sequence.Length - 1
                Dim r As Integer = rnd.Next(i, sequence.Length)
                Dim tmp As Integer = sequence(r)
                sequence(r) = sequence(i)
                sequence(i) = tmp
            Next i
        End Sub

        Private Function [Error](ByVal trainData()() As Double, ByVal weights() As Double) As Double
            ' mean squared error using supplied weights
            Dim yIndex As Integer = trainData(0).Length - 1 ' y-value (0/1) is last column
            Dim sumSquaredError As Double = 0.0
            For i = 0 To trainData.Length - 1 ' each data
                Dim computed As Double = ComputeOutput(trainData(i), weights)
                Dim desired As Double = trainData(i)(yIndex) ' ex: 0.0 or 1.0
                sumSquaredError += (computed - desired) * (computed - desired)
            Next i
            Return sumSquaredError / trainData.Length
        End Function

        Private Function ComputeOutput(ByVal dataItem() As Double, ByVal weights() As Double) As Double
            Dim z As Double = 0.0
            z += weights(0) ' the b0 constant
            For i As Integer = 0 To weights.Length - 2 ' data might include Y
                z += (weights(i + 1) * dataItem(i)) ' skip first weight
            Next i
            Return 1.0 / (1.0 + Math.Exp(-z))
        End Function

        Private Function ComputeDependent(ByVal dataItem() As Double, ByVal weights() As Double) As Integer
            Dim y As Double = ComputeOutput(dataItem, weights) ' 0.0 to 1.0
            If y <= 0.5 Then
                Return 0
            Else
                Return 1
            End If
        End Function

        Public Function Accuracy(ByVal trainData()() As Double, ByVal weights() As Double) As Double
            Dim numCorrect As Integer = 0
            Dim numWrong As Integer = 0
            Dim yIndex As Integer = trainData(0).Length - 1
            For i = 0 To trainData.Length - 1
                Dim computed As Integer = ComputeDependent(trainData(i), weights)
                Dim target As Integer = CInt(Fix(trainData(i)(yIndex))) ' risky?

                If computed = target Then
                    numCorrect += 1
                Else
                    numWrong += 1
                End If
            Next i
            Return (numCorrect * 1.0) / (numWrong + numCorrect)
        End Function
    End Class ' LogisticClassifier
End Namespace ' ns